'use strict';

var config = require('../../config.json');

var findComment = require('../lib/find-comment');
var canInjectIntoFunction = require('../lib/check-function');
var parseJsDoc = require('../lib/parse-jsdoc');
var normalizeFunctionBody = require('../lib/normalize-function-body');
var insertParametersCheck = require('../lib/insert-parameters-check');
var returnStatementVisitorFactory = require('./return-statement');

var helpers = {
    /**
     * @param {Array<NodePath>} declarations
     * @returns {NodePath|null}
     */
    declarationIsFunction: function declarationIsFunction(declarations) {
        var functionNodeDescriptor = 'init';
        var declaration = void 0;

        for (var index = 0, count = declarations.length; index < count; index++) {
            declaration = declarations[index].get(functionNodeDescriptor);

            if (declaration.isFunction()) {
                return declaration;
            }
        }

        return null;
    },


    /**
     * @param {Array<NodePath>} declarations
     * @returns {NodePath|null}
     */
    declarationIsClassExpression: function declarationIsClassExpression(declarations) {
        var classNodeDescriptor = 'init';
        var declaration = void 0;

        for (var index = 0, count = declarations.length; index < count; index++) {
            declaration = declarations[index].get(classNodeDescriptor);

            if (declaration.isClassExpression()) {
                return declaration;
            }
        }

        return null;
    },


    /**
     * @param {NodePath} path
     * @returns {Boolean}
     */
    pathIsConstructor: function pathIsConstructor(path) {
        return path.node.kind === 'constructor';
    },


    /**
     * @param {NodePath} path
     * @returns {Boolean}
     */
    pathIsExpression: function pathIsExpression(path) {
        return path.isExpressionStatement();
    },


    /**
     * @param {NodePath} path
     * @returns {Boolean}
     */
    pathIsClassDeclaration: function pathIsClassDeclaration(path) {
        return path.isClassDeclaration();
    },


    /**
     * @param {NodePath} classMethod
     * @param {NodePath} classExpression
     * @return {String}
     */
    createClassMethodName: function createClassMethodName(classMethod, classExpression) {
        var node = classMethod.node;
        var className = classExpression.node.id ? classExpression.node.id.name : config.defaultClassName;
        var kind = node.kind === 'constructor' ? '' : node.kind + ' ';

        return '' + (node.static ? 'static ' : '') + kind + className + '.' + node.key.name;
    }
};

/**
 * @param {Function} typecheckFunctionCall
 * @param {Object} globalState
 * @param {Object} t
 * @returns {Object}
 */
module.exports = function (typecheckFunctionCall, globalState, t) {
    var returnTypecheckVisitor = returnStatementVisitorFactory(typecheckFunctionCall, globalState, t);

    /**
     * @param {String} comment
     * @param {NodePath} path
     * @param {String} [name]
     */
    function executeFunctionTransformation(comment, path, name) {
        var jsDoc = parseJsDoc(comment);

        if (!jsDoc) return;

        var functionName = name || jsDoc.name || (path.node.id ? path.node.id.name : config.defaultFunctionName);

        normalizeFunctionBody(path);

        if (canInjectIntoFunction(path)) {
            insertParametersCheck(functionName, typecheckFunctionCall, path, jsDoc, globalState.useStrict, t);

            if (jsDoc.returnStatement || globalState.useStrict) {
                path.traverse(returnTypecheckVisitor, {
                    jsDoc: jsDoc,
                    functionPath: path,
                    functionName: functionName
                });
            }

            globalState.shouldInjectHelperFunction = true;
        }
    }

    /**
     * @param {NodePath} classDeclaration
     * @param {PluginPass} state
     */
    function executeTraverseForNestedClass(classDeclaration, state) {
        classDeclaration.traverse({
            ClassMethod: function ClassMethod(classPath) {
                var comment = findComment(classPath, state, globalState.hasGlobalDirective);

                if (!comment) return;

                executeFunctionTransformation(comment, classPath, helpers.createClassMethodName(classPath, classDeclaration));
            }
        });
    }

    return {
        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        VariableDeclaration: function VariableDeclaration(path, state) {
            var declarations = path.get('declarations');
            var functionDeclaration = helpers.declarationIsFunction(declarations);

            if (functionDeclaration) {
                var comment = findComment(path, state, globalState.hasGlobalDirective);

                if (comment) {
                    executeFunctionTransformation(comment, functionDeclaration);
                }
            } else {
                var classExpressionDeclaration = helpers.declarationIsClassExpression(declarations);

                if (classExpressionDeclaration) {
                    executeTraverseForNestedClass(classExpressionDeclaration, state);
                }
            }
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ReturnStatement: function ReturnStatement(path, state) {
            var argument = path.get('argument');

            if (!argument || !argument.isFunction()) return;

            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) return;

            executeFunctionTransformation(comment, argument);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        AssignmentExpression: function AssignmentExpression(path, state) {
            var rightPath = path.get('right');

            if (!rightPath.isFunction()) return;

            var expression = path.find(helpers.pathIsExpression);
            var comment = findComment(expression, state, globalState.hasGlobalDirective);

            if (!comment) return;

            executeFunctionTransformation(comment, rightPath);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ObjectProperty: function ObjectProperty(path, state) {
            var value = path.get('value');

            if (!value.isFunction()) return;

            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) return;

            executeFunctionTransformation(comment, value, path.node.key.name);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ClassMethod: function ClassMethod(path, state) {
            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) return;

            var classDeclaration = path.find(helpers.pathIsClassDeclaration);

            if (!classDeclaration) return;

            executeFunctionTransformation(comment, path, helpers.createClassMethodName(path, classDeclaration));
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        ObjectMethod: function ObjectMethod(path, state) {
            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) return;

            executeFunctionTransformation(comment, path, path.node.key.name);
        },


        /**
         * @param {NodePath} path
         * @param {PluginPass} state
         */
        'FunctionDeclaration|ArrowFunctionExpression': function FunctionDeclarationArrowFunctionExpression(path, state) {
            var comment = findComment(path, state, globalState.hasGlobalDirective);

            if (!comment) return;

            executeFunctionTransformation(comment, path);
        }
    };
};